package engine;

public class Constants {
	
	final static public int nCols = 8;
	final static public int nRows = 6;
	
	/**
	 * Eligibility = lambda of TD(lambda), which is the eligibility trace constant that determines how far back we track steps to reward
	 */
	final static public double eligibility = 0.5;
	
	/**
	 * CutOff is where we no longer continue to trace back eligibility, since it takes a long time, and won't have much of an effect on the final value
	 * (remember that this value is multiplied by the learning factor as well, so the total effect can be much smaller than this value)
	 * We trace back N steps, where N = max {n : eligibility^n > cutOff } = ceil( ln(cutOff)/ln(eligibility) ), so if we're implementing this as an optimisation
	 * we should calculate cutOff such that N is a reasonably small value (e.g. N <= 7)
	 */
	final static public double cutOff = 0.002;
	final static public int eligibilityTraceLength = (int) Math.floor( Math.log(cutOff)/Math.log(eligibility) );

	/**
	 * LearningRate, the amount we take into account each reward given
	 */
	//not constant, based on http://en.wikipedia.org/wiki/Q-learning#Learning_rate
	//final private static double learningRate = 0.1;
	static public double getLearningRate(int t, int fn) {
		return .0001 + .001*(1-getTrust(t, fn));
	}
	
	final public static double discount = 0.1;

	final public static int minDepthLearning = 1; // when trust=0
	final public static int maxDepthLearning = 4; // when trust=1
	final public static int maxDepthPlaying = 8; // when in actual production gameplay
	
	/**
	 * The epsilon in epsilon-greedy
	 */
	final static public double epsilon = 0.1;
	
	final static public int winScore = 1, loseScore = -1, tieScore = 0;
	
	/**
	 * When we raise the scores to a power (to make them positive), we must "normalise" the exponent to maintain some sense of linearity.
	 * For example with scores of 20,80 I would expect, linearly, to get 20 20% of the time and 80 80% of the time, approx. However when we
	 * raise to the power of e, we get a relation that is orders of magnitude different, which is essentially 80 100% of the time.
	 * this is because 80=20*4 so we take the result of e^20 and are essentially raising it to the power of 4, which makes it much more likely.
	 * One solution is to just shift the numbers so they're all positive.
	 * Another is to squeeze the scores linearly, and raise them to a "reasonable" power, which we ensure (because of the squeeze we guarantee
	 * the numbers lie in the interval [a,b]) is reasonable by choosing a base where base^b is sufficiently small, say 100.
	 */
	final static public double maximalExponent = 5;
	
	static public int outputInterval = 10000;
	static public int updateTInterval = 1000; // we don't need to update it every iteration, and the difference between 1 and 5000 is around 300ms/10000games
	static public double lastTrust, lastT = -updateTInterval;
	
	/**
	 * @return a value representing how much the agent trusts its current knowledge.
	 * A number in the range [0,1], 0 - no learned knowledge ==> no trust. 1 - we have converged ==> complete trust.
	 */
	static public double getTrust(int t, int fn) {
		if (t-lastT>updateTInterval) {
			lastT = t;
			switch (fn) {
			case 0:
				lastTrust = Math.min(Math.sqrt(t) / 900, 1);
				break;
			case 1:
				lastTrust = Math.min(t / 800000.0, 1);
				break;
			case 2:
				lastTrust = Math.min((double)t*t / 640000000000.0, 1);
				break;
			}
		}
		return lastTrust;
	}

	/**
	 * Number of games to play
	 */
	static public int convergedT = 1000001;
	
	/**
	 * Saves weights after the indicated number of games (used in training mode)
	 */
	static public int trainAllSaveInterval = 100000;
	
	static public int singlesGameSizeCountCutoff = nCols*nRows/4;
	
	/**
	 * Number of games to play for statistics (used in testing mode)
	 */
	static public final int numTestsForWinningStatistics = 500;
	
}
