package guiAgent;

import java.util.Arrays;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseTrackAdapter;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;

public class GUI {

	private final int width = 827, height = 710;
	
	private Image empty;
	private Image[] empties, pieces, shadows;
	
	private Label[][] holders;
	private Label message;
	public Shell shell;
	private Display display;
	
	private Role player = Role.RED;//set default color as red
	
	private final GUIPlayer agent;
	
	private ExecutorService thread = Executors.newSingleThreadExecutor();
	
	
	/**
	 * Constructor that opens a new GIU thread
	 */
	public GUI(GUIPlayer a) {

		this.agent = a;
		holders = new Label[agent.getNCols()][agent.getNRows()];

		thread.execute(new Runnable() {
			@Override
			public void run() {

				display = new Display();
				
				empty = new Image(display, "resources\\empty.png");
				empties = new Image[]{empty};
				pieces = new Image[]{new Image(display, "resources\\red.png"), new Image(display, "resources\\yellow.png") };
				shadows = new Image[]{new Image(display, "resources\\red_shadow.png"), new Image(display, "resources\\yellow_shadow.png") };
				
				init();
			}
		});
		
	}

	/**
	 * Instantiates the SWT screen with click/hover listeners
	 */
	private void init() {

		shell = new Shell(display, SWT.DIALOG_TRIM);

		shell.setText("Connect Four");
		shell.setBackground(display.getSystemColor(SWT.COLOR_WHITE));
		shell.setBackgroundMode(SWT.INHERIT_FORCE);
		shell.setImage(new Image(display, "resources\\icon.png"));
		shell.setSize(width, height);

		// center on screen
		Rectangle screen = display.getPrimaryMonitor().getBounds(), window = shell.getBounds();
		shell.setLocation(screen.x + (screen.width - window.width)/2, screen.y + (screen.height - window.height)/2);

		shell.setLayout(new GridLayout(1,true));
		
		Font font = new Font(display,"Arial",20,SWT.BOLD);
		message = new Label(shell, SWT.NONE);
		message.setFont(font);
		message.setForeground(display.getSystemColor(SWT.COLOR_BLUE));

		final Composite grid = new Composite(shell, SWT.NONE);
		GridLayout gridLayout = new GridLayout(agent.getNCols(), true);
		gridLayout.horizontalSpacing = gridLayout.verticalSpacing = 0;
		grid.setLayout(gridLayout);

		// draw grid
		for(int i=0; i<agent.getNCols(); ++i) {
			final Composite column = new Composite(grid, SWT.NONE);
			RowLayout rowLayout = new RowLayout(SWT.VERTICAL);
			rowLayout.spacing = rowLayout.marginTop = rowLayout.marginLeft = rowLayout.marginRight = rowLayout.marginBottom = 0;
			column.setLayout(rowLayout);

			final int _i = i;
			for(int j=0; j<agent.getNRows(); ++j) {
				holders[i][j] = new Label(column, SWT.NONE);
				holders[i][j].setImage(empty);
				// would prefer the composite, but the event doesn't bubble in SWT
				holders[i][j].addMouseTrackListener(new MouseTrackAdapter() {
					@Override
					public void mouseEnter(MouseEvent e) {
						// max, so that when we're not in the game (i.e. player==-1) we don't hit an out of bounds exception
						switchFirstThatEquals(_i, empties, shadows[player.ordinal()]);
					}
					@Override
					public void mouseExit(MouseEvent e) {
						switchFirstThatEquals(_i, shadows, empty);
					}
				});
				holders[i][j].addMouseListener(new MouseAdapter() {
					@Override
					public void mouseDown(MouseEvent e) {
						
						if (!agent.isInGame()) {
							alert("The game hasn't started yet");
							return;
						}
						
						if (!agent.isMyTurn()) {
							alert("Still waiting for the opponent's move");
							return;
						}

						if (getFirstThatEquals(_i, shadows) < 0) {
							alert("Illegal move. Try again");
							return;
						}

						agent.setMove(_i+1);				
						
						alert("Waiting for the opponent's move");
					}
				});
			}
		}

		shell.open();

		while (!shell.isDisposed()) {
			if (!display.readAndDispatch())
				display.sleep();
		}
		
		display.dispose();
		font.dispose();
		empty.dispose();
		pieces[0].dispose();
		pieces[1].dispose();
		shadows[0].dispose();
		shadows[1].dispose();
		
		thread.shutdownNow();
		
	}
	
	/**
	 * Clear the pieces from the board
	 */
	public void start(boolean isRed) {

		player = isRed? Role.RED:Role.YELLOW;

		try { // ugly hack to wait for the display to be ready without using proper event signalling
			while(display==null)
				Thread.sleep(100);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		
		display.asyncExec(new Runnable() {
			@Override
			public void run() {

				//if(shell.isDisposed()) init();
								
				for(int i=0; i<agent.getNCols(); ++i)
					for(int j=0;j<agent.getNRows(); ++j)
						holders[i][j].setImage(empty);
				alert("New game");
				
			}
		});
		
	}
	
	/**
	 * Display the winner
	 */
	public void finish(final int score) {
		
		display.asyncExec(new Runnable() {
			@Override
			public void run() {
			
				//if(shell.isDisposed()) return;

				alert("The game finished. You " + (score>0 ? "won!":"lost."));
			}
		});
	}

	/**
	 * Draws a move on the board
	 */
	public void move(int move, final boolean isOpponent) {
		
		final int _move = move-1; // the indices are 0-based
		display.asyncExec(new Runnable() {
			@Override
			public void run() {
			
				//if(shell.isDisposed()) return;
				
				boolean wasHover = switchFirstThatEquals(_move, shadows, empty);
				
				switchFirstThatEquals(_move, empties, pieces[isOpponent ? player.getOpponent().ordinal() : player.ordinal()]);
				
				if (wasHover) // replace the hover
					switchFirstThatEquals(_move, empties, shadows[player.ordinal()]) ;
				
				alert(isOpponent? "Your move" : "Waiting for the opponent's move");
			}
		});
	}
	
	/**
	 * 
	 * @param col           index of the column
	 * @param compare       array of images to compare to
	 * @param replaceWith   new image to replace with
	 * @return true if it found and replaced an image from within the array, while searching from the bottom upwards!
	 */
	private boolean switchFirstThatEquals(int col, Image[] compare, Image replaceWith) {
		int j = getFirstThatEquals(col, compare);
		if (j>=0) { // found
			holders[col][j].setImage(replaceWith);
			return true;
		}
		return false;
	}
	
	/**
	 * 
	 * @param col           index of the column
	 * @param compare       array of images to compare to
	 * @return the index of the row that contains one of the images in the compare array, while searching from the bottom upwards!
	 */
	private int getFirstThatEquals(int col, Image[] compare) {
		int j;
		for(j=agent.getNRows()-1; j>=0 && !Arrays.asList(compare).contains(holders[col][j].getImage()); --j);
		return j;
	}
	
	/**
	 * Displays an alert/message to the user on the screen (on an SWT label)
	 */
	private void alert(String msg) {
		message.setText(msg);
		message.getParent().layout();
	}

	public void dispose() {
		if (display.isDisposed()||shell.isDisposed()) return;
		display.asyncExec(new Runnable() {
			@Override
			public void run() {
				shell.dispose();
			}
		});
	}
	
}
