package engine;

import runtime.IPlayer;
import runtime.Manager;
import board.Board;


public class Sarsa implements IPlayer {

	/**
	 * t holds the time, i.e. the number of steps/iterations/games we've played since learning began
	 */
	private int t = 1;
	
	/**
	 * nextMoveDecider holds the explore-exploit policy used. can be epsilon greedy or stochastic in our implementation
	 */
	private ExploitExplore nextMoveDecider = new DeciderEpsilonGreedy(Constants.minDepthLearning, Constants.maxDepthLearning);
	
	/**
	 * The vector of weights
	 */
	private double[] weights;
	
	/**
	 * Number of features
	 */
	final private int numFeatures;
	
	/**
	 * A board representing the game state
	 */
	private Board board;
	
	/**
	 * A boolean value indicating the agent is the first player
	 */
	private boolean didIStart;
	
	/**
	 * Trust function to use (we tested three options)
	 */
	private int trustFunction;
	
	public Sarsa(String board, int trust) {
		this.board = Manager.getBoard(board);
		numFeatures = this.board.getFeatureVector().length;
		weights = new double[numFeatures];
		trustFunction = trust;
	}
	
	@Override
	public int getMove() {
		
		// CHOOSE MOVE FROM CURRENT KNOWLEDGE

		int move = nextMoveDecider.getNextMove(board, weights, Constants.getTrust(t, trustFunction));

		board.setMove(move, false);
		
		return move;
	}
	
	/**
	 * Calculates the score of the new state
	 * @param weights - vector of weights
	 * @param featuresVector
	 * @return score of the board state
	 */
	static double calculateScore(double[] weights, double[] featuresVector) {
		double score = 0;
		for(int i=0; i<featuresVector.length; ++i) {
			score += weights[i] * featuresVector[i];
		}
		return score;
	}

	@Override
	public void setOpponentsMove(int move) {
		board.setMove(move, true);
	}

	@Override
	public void start(boolean doIStart) {
		didIStart = doIStart;
		board.clear();
	}

	@Override
	public void finish(int winner) {
		
		// UPDATE WEIGHTS WITH ENVIRONMENT FEEDBACK

		updateWeights(winner==2? Constants.tieScore : winner==0 ? Constants.winScore : Constants.loseScore);
		
		++t;
	}

	/**
	 * Updates the weight vector according to the environment feedback
	 * @param score - game result. can be tieScore, winScore or loseScore
	 */
	private void updateWeights(int score) {

		double learningRate = Constants.getLearningRate(t, trustFunction);

		double scoreFinal = Sarsa.calculateScore(weights, board.getFeatureVector());
		int lastMove = board.popMove();
		double scoreBefore = Sarsa.calculateScore(weights, board.getFeatureVector());
		board.setMove(lastMove, score==Constants.tieScore? didIStart : score==Constants.loseScore);

		double[] eligibility = calculateEligibilityVector();
		
		// delta ~ new evaluation - old evaluation, i.e. TD
		// Off policy
		double delta = score + Constants.discount*scoreFinal - scoreBefore;
		
		for(int i=0; i<weights.length; ++i) {
			weights[i] += learningRate * delta * eligibility[i];
		}

	}

	/**
	* Based on Alppaydin, 2010, pages 462-3
	* @return vector of eligibilities as the eligibility trace is also taken into account
	**/
	private double[] calculateEligibilityVector() {
		
		double[] e = new double[numFeatures];
		
		double eligibility = 1;
		
		for(int i=0; i<Constants.eligibilityTraceLength; ++i) {
			
			double[] featuresV = board.getFeatureVector();
			
			for(int j=0; j<numFeatures; ++j) {
				e[j] += eligibility * featuresV[j];
			}
			
			if (board.popMove() == -1) break;
			
			eligibility *= Constants.eligibility * Constants.discount;
		}
				
		return e;
	}

	/**
	* @return true if we reached to the number of games set 
	**/
	public boolean hasConverged() {
		return t>Constants.convergedT;
	}
	
	public int getIterations() {
		return t;
	}
	
	public double[] getWeights() {
		return weights;
	}
	
	public int getTrustFn() {
		return trustFunction;
	}

}
